# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections import defaultdict
from typing import Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType, get_output_types, recursive_tile
from ...core.operand import OperandStage
from ...dataframe.core import DATAFRAME_TYPE
from ...dataframe.utils import parse_index
from ...serialization.serializables import AnyField, BoolField, \
    Int64Field, Float32Field, TupleField, ReferenceField, FieldTypes
from ...tensor.core import TENSOR_CHUNK_TYPE
from ...tensor.random import RandomStateField
from ...tensor.utils import gen_random_seeds
from ...typing import TileableType
from ...utils import has_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin, LearnShuffleProxy
from ..utils.shuffle import LearnShuffle


def _extract_bagging_io(io_list: Iterable, op: LearnOperand,
                        output: bool = False):
    if not isinstance(io_list, Iterable):
        io_list = [io_list]
    input_iter = iter(io_list)
    out = [
        next(input_iter),
        next(input_iter) if op.with_labels else None,
        next(input_iter) if op.with_weights else None,
        next(input_iter) if output and op.with_feature_indices else None,
    ]
    return out


def _get_by_iloc(x, idx, axis=0):
    if hasattr(x, 'iloc'):
        item_getter = x.iloc
    else:
        item_getter = x
    if axis == 0:
        return item_getter[idx]
    else:
        return item_getter[:, idx]


def _concat_on_axis(data_list, axis=0, out_chunk=None):
    if isinstance(out_chunk, TENSOR_CHUNK_TYPE):
        return np.concatenate(data_list, axis=axis)
    else:
        return pd.concat(data_list, axis=axis)


def _concat_by_row(row, out_chunk=None):
    arr = np.empty((1,), dtype=object)
    arr[0] = _concat_on_axis(row.tolist(), axis=0, out_chunk=out_chunk)
    return arr


class BaggingSample(LearnShuffle, LearnOperandMixin):
    _op_type_ = opcodes.BAGGING_SHUFFLE_SAMPLE

    n_estimators: int = Int64Field('n_estimators')
    max_samples = AnyField('max_samples')
    max_features = AnyField('max_features')
    bootstrap: bool = BoolField('bootstrap')
    bootstrap_features: bool = BoolField('bootstrap_features')

    random_state = RandomStateField('random_state')
    sample_random_state = RandomStateField('sample_random_state')
    feature_random_state = RandomStateField('feature_random_state')

    reducer_ratio: float = Float32Field('reducer_ratio')
    n_reducers: int = Int64Field('n_reducers', default=None)
    column_offset: int = Int64Field('column_offset', default=None)

    chunk_shape: Tuple[int] = TupleField('chunk_shape', FieldTypes.int64)
    with_labels: bool = BoolField('with_labels')
    with_weights: bool = BoolField('with_weights')
    with_feature_indices: bool = BoolField('with_feature_indices')

    def __init__(self, max_samples: Union[int, float] = 1.0,
                 max_features: Union[int, float] = 1.0,
                 bootstrap: bool = True,
                 bootstrap_features: bool = False,
                 random_state: np.random.RandomState = None,
                 reducer_ratio: float = 1.0, **kw):
        super().__init__(
            bootstrap=bootstrap,
            bootstrap_features=bootstrap_features,
            max_samples=max_samples,
            max_features=max_features,
            reducer_ratio=reducer_ratio,
            random_state=random_state,
            **kw
        )
        if self.random_state is None:
            self.random_state = np.random.RandomState()

    @property
    def output_limit(self) -> int:
        if self.with_feature_indices and self.stage != OperandStage.map:
            return 1 + self.with_labels + self.with_weights + self.with_feature_indices
        return 1

    def __call__(self, in_sample: TileableType,
                 in_labels: Optional[TileableType] = None,
                 in_weights: Optional[TileableType] = None):
        self._output_types = get_output_types(in_sample, in_labels, in_weights)

        self.with_labels = in_labels is not None
        self.with_weights = in_weights is not None
        axis_keep_shape = [
            isinstance(self.max_samples, float) and self.max_samples == 1.0,
            isinstance(self.max_features, float) and self.max_features == 1.0,
        ]
        self.with_feature_indices = not axis_keep_shape[1] or self.bootstrap_features
        if self.with_feature_indices:
            self._output_types += (OutputType.tensor,)

        new_shape = tuple(s if keep_shape else np.nan
                          for s, keep_shape in zip(in_sample.shape, axis_keep_shape))

        kws = []

        data_params = in_sample.params
        data_params['shape'] = new_shape
        kws.append(data_params)

        if in_labels is not None:
            labels_params = in_labels.params
            labels_params['shape'] = (new_shape[0],)
            kws.append(labels_params)

        if in_weights is not None:
            weights_params = in_weights.params
            weights_params['shape'] = (new_shape[0],)
            kws.append(weights_params)

        if self.with_feature_indices:
            feature_params = {
                'shape': (self.n_estimators, new_shape[1]),
                'dtype': np.dtype(int),
            }
            kws.append(feature_params)

        inputs = [in_sample]
        if in_labels is not None:
            inputs.append(in_labels)
        if in_weights is not None:
            inputs.append(in_weights)

        return self.new_tileables(inputs, kws=kws)

    @classmethod
    def _scatter_samples(cls, max_samples: Union[int, float], nsplits: Tuple[int],
                         random_state: np.random.RandomState, n_estimators: int) -> np.ndarray:
        nsp_array = np.array(nsplits)
        dim_size = nsp_array.sum()
        if isinstance(max_samples, int):
            expect_sample_count = max_samples
        else:
            expect_sample_count = int(max_samples * nsp_array.sum())

        if expect_sample_count == dim_size:
            return np.array([list(nsplits)] * n_estimators)

        split_probs = nsp_array / dim_size
        return random_state.multinomial(expect_sample_count, split_probs, size=n_estimators)

    @classmethod
    def tile(cls, op: 'BaggingSample'):
        in_sample, in_labels, in_weights, _ = _extract_bagging_io(op.inputs, op, output=False)
        out_data, out_labels, out_weights, out_feature_indices = _extract_bagging_io(
            op.outputs, op, output=True)

        # make sure all shapes are computed
        if has_unknown_shape(in_sample) or \
                (in_labels is not None and has_unknown_shape(in_labels)) or \
                (in_weights is not None and has_unknown_shape(in_weights)):
            yield

        to_tile = []
        if in_labels is not None:
            in_labels = in_labels.rechunk({0: in_sample.nsplits[0]})
            to_tile.append(in_labels)
        if in_weights is not None:
            in_weights = in_weights.rechunk({0: in_sample.nsplits[0]})
            to_tile.append(in_weights)

        # tile rechunks
        if to_tile:
            tiled = yield from recursive_tile(*to_tile)
            tiled_iter = iter(tiled)
            if in_labels is not None:
                in_labels = next(tiled_iter)
            if in_weights is not None:
                in_weights = next(tiled_iter)

        random_seeds = [gen_random_seeds(n, op.random_state)
                        for n in in_sample.chunk_shape]

        axis_keep_shape = [
            isinstance(op.max_samples, float) and op.max_samples == 1.0
            and not op.bootstrap,
            isinstance(op.max_features, float) and op.max_features == 1.0
            and not op.bootstrap_features,
        ]

        n_reducers = op.n_reducers if op.n_reducers is not None \
            else max(1, int(in_sample.chunk_shape[0] * op.reducer_ratio))

        # todo implement sampling without replacements
        map_chunks = []
        max_samples_splits = cls._scatter_samples(op.max_samples, in_sample.nsplits[0],
                                                  op.random_state, op.n_estimators)
        max_features_splits = cls._scatter_samples(op.max_features, in_sample.nsplits[1],
                                                   op.random_state, op.n_estimators)

        column_cum_offset = np.concatenate([[0], np.cumsum(in_sample.nsplits[1])])
        for chunk in in_sample.chunks:
            new_op = op.copy().reset_key()
            new_op.random_state = None
            new_op.sample_random_state = np.random.RandomState(random_seeds[0][chunk.index[0]])
            new_op.feature_random_state = np.random.RandomState(random_seeds[1][chunk.index[1]])
            new_op.stage = OperandStage.map
            new_op.max_samples = max_samples_splits[:, chunk.index[0]]
            new_op.max_features = max_features_splits[:, chunk.index[1]]
            new_op.n_reducers = n_reducers
            new_op.column_offset = int(column_cum_offset[chunk.index[1]])

            if chunk.index[0] != 0:
                new_op.with_feature_indices = False

            if chunk.index[1] != in_sample.chunk_shape[1] - 1:
                new_op.with_weights = False
                new_op.with_labels = False

            params = chunk.params
            params['shape'] = tuple(s if keep_shape else np.nan
                                    for s, keep_shape in zip(chunk.shape, axis_keep_shape))

            input_chunks = [chunk]
            if new_op.with_labels:
                input_chunks.append(in_labels.cix[chunk.index[0]])
            if new_op.with_weights:
                input_chunks.append(in_weights.cix[chunk.index[0]])
            map_chunks.append(new_op.new_chunk(input_chunks, **params))

        shuffle_op = LearnShuffleProxy(output_types=[OutputType.tensor]) \
            .new_chunk(map_chunks, dtype=np.dtype(int), shape=())

        remain_reducers = op.n_estimators % n_reducers
        reduce_data_chunks = []
        reduce_labels_chunks = []
        reduce_weights_chunks = []
        reduce_feature_chunks = []
        for idx in range(n_reducers):
            new_op = op.copy().reset_key()
            new_op.random_state = None
            new_op.stage = OperandStage.reduce
            new_op.chunk_shape = in_sample.chunk_shape
            new_op.n_estimators = op.n_estimators // n_reducers
            if remain_reducers:
                remain_reducers -= 1
                new_op.n_estimators += 1

            if new_op.n_estimators == 0:
                continue

            kws = []

            data_params = out_data.params
            data_params['index'] = (idx, 0)
            data_params['shape'] = (np.nan, out_data.shape[1])
            kws.append(data_params)

            if op.with_labels:
                labels_params = out_labels.params
                labels_params['index'] = (idx,)
                labels_params['shape'] = (np.nan,)
                kws.append(labels_params)

            if op.with_weights:
                weights_params = out_weights.params
                weights_params['index'] = (idx,)
                weights_params['shape'] = (np.nan,)
                kws.append(weights_params)

            if op.with_feature_indices:
                feature_params = {
                    'index': (idx, 0),
                    'shape': (new_op.n_estimators, out_feature_indices.shape[1]),
                    'dtype': np.dtype(int),
                }
                kws.append(feature_params)

            chunks = new_op.new_chunks([shuffle_op], kws=kws)
            data_chunk, labels_chunk, weights_chunk, feature_chunk \
                = _extract_bagging_io(chunks, op, output=True)

            reduce_data_chunks.append(data_chunk)
            if labels_chunk is not None:
                reduce_labels_chunks.append(labels_chunk)
            if weights_chunk is not None:
                reduce_weights_chunks.append(weights_chunk)
            if feature_chunk is not None:
                reduce_feature_chunks.append(feature_chunk)

        new_op = op.copy().reset_key()

        kws = [{
            'chunks': reduce_data_chunks,
            'nsplits': ((np.nan,) * len(reduce_data_chunks), (out_data.shape[1],)),
            **out_data.params
        }]
        if op.with_labels:
            kws.append({
                'chunks': reduce_labels_chunks,
                'nsplits': ((np.nan,) * len(reduce_data_chunks),),
                **out_labels.params
            })
        if op.with_weights:
            kws.append({
                'chunks': reduce_weights_chunks,
                'nsplits': ((np.nan,) * len(reduce_data_chunks),),
                **out_weights.params
            })
        if op.with_feature_indices:
            estimator_nsplit = tuple(c.op.n_estimators for c in reduce_data_chunks)
            kws.append({
                'chunks': reduce_feature_chunks,
                'nsplits': (estimator_nsplit, (out_feature_indices.shape[1],)),
                **out_feature_indices.params
            })
        return new_op.new_tileables(op.inputs, kws=kws)

    @classmethod
    def _gen_sample_indices(cls, max_range: int, size: int,
                            random_state: np.random.RandomState,
                            with_replacement: bool = False):
        if not with_replacement:
            result = random_state.choice(np.arange(max_range), size, False)
        else:
            result = random_state.randint(0, max_range - 1, size)
        result.sort()
        return result

    @classmethod
    def _execute_map(cls, ctx, op: 'BaggingSample'):
        in_sample, in_labels, in_weights, _ = _extract_bagging_io(op.inputs, op, output=False)
        in_sample_data = ctx[in_sample.key]
        in_labels_data = ctx[in_labels.key] if op.with_labels else None
        in_weights_data = ctx[in_weights.key] if op.with_weights else None
        out_samples = op.outputs[0]

        remains = op.n_estimators % op.n_reducers
        reducer_iters = [itertools.repeat(idx, 1 + op.n_estimators // op.n_reducers)
                         for idx in range(remains)]
        reducer_iters += [itertools.repeat(idx, op.n_estimators // op.n_reducers)
                          for idx in range(remains, op.n_reducers)]
        reducer_iter = itertools.chain(*reducer_iters)

        result_store = defaultdict(lambda: ([], [], [], []))
        for est_id in range(op.n_estimators):
            sampled_data = in_sample_data
            sampled_labels = in_labels_data
            sampled_weights = in_weights_data

            if op.max_samples[est_id] != in_sample_data.shape[0]:
                sample_indices = cls._gen_sample_indices(
                    in_sample_data.shape[0], op.max_samples[est_id],
                    op.sample_random_state, op.bootstrap)

                sampled_data = _get_by_iloc(sampled_data, sample_indices)
                if sampled_labels is not None:
                    sampled_labels = _get_by_iloc(sampled_labels, sample_indices)
                if sampled_weights is not None:
                    sampled_weights = _get_by_iloc(sampled_weights, sample_indices)

            if op.max_features[est_id] != in_sample_data.shape[1]:
                feature_indices = cls._gen_sample_indices(
                    in_sample_data.shape[1], op.max_features[est_id],
                    op.feature_random_state, op.bootstrap_features)

                sampled_data = _get_by_iloc(sampled_data, feature_indices, axis=1)
                if not op.with_feature_indices:
                    feature_indices = None
            else:
                feature_indices = None

            samples, labels, weights, feature_idx_array \
                = result_store[next(reducer_iter)]
            samples.append(sampled_data)
            if sampled_labels is not None:
                labels.append(sampled_labels)
            if sampled_weights is not None:
                weights.append(sampled_weights)
            if feature_indices is not None:
                feature_idx_array.append(feature_indices + op.column_offset)

        for reducer_id, (samples, labels, weights, feature_idx_array) \
                in result_store.items():
            ctx[out_samples.key, (reducer_id, 0)] = \
                tuple(samples + labels + weights + feature_idx_array)

    @classmethod
    def _execute_reduce(cls, ctx, op: 'BaggingSample'):
        out_data, out_labels, out_weights, out_feature_indices = _extract_bagging_io(
            op.outputs, op, output=True)

        input_keys = op.inputs[0].op.source_keys
        input_idxes = op.inputs[0].op.source_idxes

        sample_holder = [
            np.empty(op.chunk_shape, dtype=object)
            for _ in range(op.n_estimators)
        ]

        labels_holder = [
            np.empty(op.chunk_shape[0], dtype=object)
            for _ in range(op.n_estimators)
        ] if op.with_labels else None

        weights_holder = [
            np.empty(op.chunk_shape[0], dtype=object)
            for _ in range(op.n_estimators)
        ] if op.with_weights else None

        feature_indices_holder = [
            np.empty(op.chunk_shape[1], dtype=object)
            for _ in range(op.n_estimators)
        ] if op.with_feature_indices else None

        for input_key, input_idx in zip(input_keys, input_idxes):
            add_feature_index = input_idx[0] == 0
            add_label_weight = input_idx[1] == op.chunk_shape[1] - 1
            chunk_data = ctx[input_key, out_data.index]

            num_groups = 1
            if add_feature_index and op.with_feature_indices:  # contains feature indices
                num_groups += 1
            if add_label_weight:  # contains label or weight
                num_groups += int(op.with_weights) + int(op.with_labels)

            sample_count = len(chunk_data) // num_groups
            assert len(chunk_data) % num_groups == 0

            group_iter = (chunk_data[i * sample_count: (i + 1) * sample_count]
                          for i in range(num_groups))

            for data_idx, sample in enumerate(next(group_iter)):
                sample_holder[data_idx][input_idx] = sample

            if add_label_weight:
                if op.with_labels:
                    for data_idx, label in enumerate(next(group_iter)):
                        labels_holder[data_idx][input_idx[0]] = label
                if op.with_weights:
                    for data_idx, weight in enumerate(next(group_iter)):
                        weights_holder[data_idx][input_idx[0]] = weight

            if add_feature_index and op.with_feature_indices:
                for data_idx, feature_index in enumerate(next(group_iter)):
                    feature_indices_holder[data_idx][input_idx[1]] = feature_index

        data_results: List[Optional[np.ndarray]] = [None] * len(sample_holder)
        for est_idx, sample_mat in enumerate(sample_holder):
            row_chunks = np.apply_along_axis(_concat_by_row, axis=0, arr=sample_mat, out_chunk=out_data)
            data_results[est_idx] = _concat_on_axis(row_chunks[0].tolist(), axis=1, out_chunk=out_data)
        ctx[out_data.key] = tuple(data_results)

        for out, holder in zip(
            (out_labels, out_weights, out_feature_indices),
            (labels_holder, weights_holder, feature_indices_holder),
        ):
            if out is None:
                continue
            results: List[Optional[np.ndarray]] = [None] * len(holder)
            for est_idx, labels_vct in enumerate(holder):
                results[est_idx] = _concat_on_axis(labels_vct.tolist(), out_chunk=out)
            if holder is feature_indices_holder:
                ctx[out.key] = np.stack(results)
            else:
                ctx[out.key] = tuple(results)

    @classmethod
    def execute(cls, ctx, op: 'BaggingSample'):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)


class BaggingSampleReindex(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.BAGGING_SHUFFLE_REINDEX

    n_estimators: int = Int64Field('n_estimators')
    start_col_index: int = Int64Field('start_col_index', 0)
    feature_indices: TileableType = ReferenceField('feature_indices', default=None)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.feature_indices is not None:
            self.feature_indices = inputs[-1]

    def __call__(self, data: TileableType, feature_indices: TileableType = None):
        self._output_types = get_output_types(data)
        inputs = [data]
        self.feature_indices = feature_indices
        params = data.params
        if feature_indices is not None:
            inputs.append(feature_indices)
            params['shape'] = (data.shape[0], np.nan)
        if isinstance(data, DATAFRAME_TYPE):
            params['index_value'] = parse_index(pd.Int64Index([]), data.key)
        return self.new_tileable(inputs, **params)

    @classmethod
    def tile(cls, op: "BaggingSampleReindex"):
        t_data = op.inputs[0]
        t_out = op.outputs[0]
        t_feature_idxes = op.feature_indices
        cum_nsplits = np.cumsum(np.concatenate([[0], t_data.nsplits[1]]))

        if t_feature_idxes is None:
            out = t_data
            if out.chunk_shape[1] > 1:
                out = yield from recursive_tile(out.rechunk({1: (out.shape[1],)}))
            return out

        # generate map chunks
        map_holder = np.empty(t_data.chunk_shape + (t_feature_idxes.chunk_shape[0],),
                              dtype=np.dtype(object))
        for chunk in t_data.chunks:
            for feature_idx_chunk in t_feature_idxes.chunks:
                new_op = op.copy().reset_key()
                new_op.stage = OperandStage.map
                new_op.start_col_index = int(cum_nsplits[chunk.index[1]])
                params = chunk.params
                new_index = params['index'] = chunk.index + (feature_idx_chunk.index[0],)
                if t_feature_idxes.chunk_shape[0] == 1:
                    new_index = new_index[:-1]
                map_holder[new_index] = new_op.new_chunk(
                    [chunk, feature_idx_chunk], **params
                )
        if op.feature_indices.chunk_shape[0] == 1:
            chunks = map_holder.reshape((t_data.chunk_shape[0],)).tolist()
        else:
            def _gen_combine_chunk(chunks):
                new_op = op.copy().reset_key()
                new_op.feature_indices = None
                new_op.stage = OperandStage.combine
                params = chunks[0].params
                params['shape'] = (chunks[0].shape[0], op.feature_indices.shape[1])
                params['index'] = (chunks[0].index[0], chunks[0].index[2])
                if isinstance(t_data, DATAFRAME_TYPE):
                    params['index_value'] = parse_index(pd.Int64Index([]), chunks[0].key)
                inputs = chunks.tolist()
                return new_op.new_chunk(inputs, **params)

            chunks_array = np.apply_along_axis(_gen_combine_chunk, 1, map_holder)
            chunks = chunks_array.reshape((chunks_array.size,)).tolist()

        new_op = op.copy().reset_key()
        new_nsplits = (t_data.nsplits[0], (op.feature_indices.shape[1],) * t_feature_idxes.chunk_shape[0])
        return new_op.new_tileables(op.inputs, chunks=chunks,
                                    nsplits=new_nsplits, **t_out.params)

    @classmethod
    def _execute_map(cls, ctx, op: "BaggingSampleReindex"):
        data = ctx[op.inputs[0].key]
        feature_idx = ctx[op.feature_indices.key] - op.start_col_index
        filtered = []
        for row in feature_idx:
            row = row[(row >= 0) & (row < data.shape[1])]
            filtered.append(_get_by_iloc(data, row, axis=1))
        ctx[op.outputs[0].key] = tuple(filtered)

    @classmethod
    def _execute_combine(cls, ctx, op: "BaggingSampleReindex"):
        data_inputs = [ctx[c.key] for c in op.inputs]
        concatenated = []
        for data_input in zip(*data_inputs):
            concatenated.append(_concat_on_axis(data_input, 1, op.inputs[0]))
        ctx[op.outputs[0].key] = tuple(concatenated)

    @classmethod
    def execute(cls, ctx, op: "BaggingSampleReindex"):
        if op.stage == OperandStage.combine:
            cls._execute_combine(ctx, op)
        else:
            cls._execute_map(ctx, op)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import re
import shutil
import subprocess
import sys
import warnings
from sysconfig import get_config_var

from pkg_resources import parse_version
from setuptools import setup, Extension, Command

import numpy as np
from Cython.Build import cythonize
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.sdist import sdist

try:
    import distutils.ccompiler
    if sys.platform != 'win32':
        from numpy.distutils.ccompiler import CCompiler_compile
        distutils.ccompiler.CCompiler.compile = CCompiler_compile
except ImportError:
    pass

# From https://github.com/pandas-dev/pandas/pull/24274:
# For mac, ensure extensions are built for macos 10.9 when compiling on a
# 10.9 system or above, overriding distuitls behaviour which is to target
# the version that python was built for. This may be overridden by setting
# MACOSX_DEPLOYMENT_TARGET before calling setup.py
if sys.platform == 'darwin':
    if 'MACOSX_DEPLOYMENT_TARGET' not in os.environ:
        target_macos_version = "10.9"
        parsed_macos_version = parse_version(target_macos_version)

        current_system = parse_version(platform.mac_ver()[0])
        python_target = parse_version(
            get_config_var('MACOSX_DEPLOYMENT_TARGET'))
        if python_target < parsed_macos_version <= current_system:
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = target_macos_version


repo_root = os.path.dirname(os.path.abspath(__file__))


def execfile(fname, globs, locs=None):
    locs = locs or globs
    exec(compile(open(fname).read(), fname, "exec"), globs, locs)


version_file_path = os.path.join(repo_root, 'mars', '_version.py')
version_ns = {'__file__': version_file_path}
execfile(version_file_path, version_ns)
version = version_ns['__version__']
# check version vs tag
if os.environ.get('GIT_TAG') and re.search(r'v\d', os.environ['GIT_TAG']) \
        and os.environ['GIT_TAG'] != 'v' + version:
    raise ValueError('Tag %r does not match source version %r'
                     % (os.environ['GIT_TAG'], version))


if os.path.exists(os.path.join(repo_root, '.git')):
    git_info = version_ns['get_git_info']()
    if git_info:
        with open(os.path.join(repo_root, 'mars', '.git-branch'), 'w') as git_file:
            git_file.write(' '.join(git_info))

cythonize_kw = dict(language_level=sys.version_info[0])
cy_extension_kw = dict()
if os.environ.get('CYTHON_TRACE'):
    cy_extension_kw['define_macros'] = [('CYTHON_TRACE_NOGIL', '1'), ('CYTHON_TRACE', '1')]
    cythonize_kw['compiler_directives'] = {'linetrace': True}

if 'MSC' in sys.version:
    extra_compile_args = ['/Ot', '/I' + os.path.join(repo_root, 'misc')]
    cy_extension_kw['extra_compile_args'] = extra_compile_args
else:
    extra_compile_args = ['-O3']
    cy_extension_kw['extra_compile_args'] = extra_compile_args


def _discover_pyx():
    exts = dict()
    for root, _, files in os.walk(os.path.join(repo_root, 'mars')):
        for fn in files:
            if not fn.endswith('.pyx'):
                continue
            full_fn = os.path.relpath(os.path.join(root, fn), repo_root)
            mod_name = full_fn.replace('.pyx', '').replace(os.path.sep, '.')
            exts[mod_name] = Extension(mod_name, [full_fn], **cy_extension_kw)
    return exts


cy_extension_kw['include_dirs'] = [np.get_include()]
extensions_dict = _discover_pyx()
cy_extensions = list(extensions_dict.values())

extensions = cythonize(cy_extensions, **cythonize_kw) + \
    [Extension('mars.lib.mmh3', ['mars/lib/mmh3_src/mmh3module.cpp', 'mars/lib/mmh3_src/MurmurHash3.cpp'])]


class ExtraCommandMixin:
    _extra_pre_commands = []

    def run(self):
        [self.run_command(cmd) for cmd in self._extra_pre_commands]
        super().run()

    @classmethod
    def register_pre_command(cls, cmd):
        cls._extra_pre_commands.append(cmd)


class CustomInstall(ExtraCommandMixin, install):
    pass


class CustomDevelop(ExtraCommandMixin, develop):
    pass


class CustomSDist(ExtraCommandMixin, sdist):
    pass


class BuildWeb(Command):
    """build_web command"""

    user_options = []
    _web_src_path = 'mars/services/web/ui'
    _web_dest_path = 'mars/services/web/static/bundle.js'
    _commands = [
        ['npm', 'install'],
        ['npm', 'run', 'bundle'],
    ]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    @classmethod
    def run(cls):
        if int(os.environ.get('NO_WEB_UI', '0')):
            return

        npm_path = shutil.which('npm')
        web_src_path = os.path.join(repo_root, *cls._web_src_path.split('/'))
        web_dest_path = os.path.join(repo_root, *cls._web_dest_path.split('/'))

        if not os.path.exists(web_src_path):
            return
        elif npm_path is None:
            if not os.path.exists(web_dest_path):
                warnings.warn('Cannot find NPM, may affect displaying Mars Web')
            return

        replacements = {'npm': npm_path}
        cmd_errored = False
        for cmd in cls._commands:
            cmd = [replacements.get(c, c) for c in cmd]
            proc_result = subprocess.run(cmd, cwd=web_src_path)
            if proc_result.returncode != 0:
                warnings.warn(f'Failed when running `{" ".join(cmd)}`')
                cmd_errored = True
                break
        if not cmd_errored:
            assert os.path.exists(cls._web_dest_path)


CustomInstall.register_pre_command('build_web')
CustomDevelop.register_pre_command('build_web')
CustomSDist.register_pre_command('build_web')


setup_options = dict(
    version=version,
    ext_modules=extensions,
    cmdclass={
        'build_web': BuildWeb,
        'install': CustomInstall,
        'develop': CustomDevelop,
        'sdist': CustomSDist
    },
)
setup(**setup_options)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Iterable, List, Optional, Tuple

from ...core import ChunkGraph, DAG
from ...serialization.serializables import Serializable, StringField, \
    ReferenceField, Int32Field, Int64Field, Float64Field, \
    BoolField, AnyField, DictField, ListField, TupleField, FieldTypes
from ...typing import BandType


class SubtaskStatus(Enum):
    pending = 0
    running = 1
    succeeded = 2
    errored = 3
    cancelled = 4

    @property
    def is_done(self) -> bool:
        return self in (SubtaskStatus.succeeded,
                        SubtaskStatus.errored,
                        SubtaskStatus.cancelled)


class Subtask(Serializable):
    subtask_id: str = StringField('subtask_id')
    subtask_name: str = StringField('subtask_name')
    session_id: str = StringField('session_id')
    task_id: str = StringField('task_id')
    chunk_graph: ChunkGraph = ReferenceField('chunk_graph', ChunkGraph)
    expect_bands: List[BandType] = ListField('expect_bands', FieldTypes.tuple)
    virtual: bool = BoolField('virtual')
    retryable: bool = BoolField('retryable')
    priority: Tuple[int, int] = TupleField('priority', FieldTypes.int32)
    rerun_time: int = Int32Field('rerun_time')
    extra_config: dict = DictField('extra_config')

    def __init__(self,
                 subtask_id: str = None,
                 session_id: str = None,
                 task_id: str = None,
                 chunk_graph: ChunkGraph = None,
                 subtask_name: str = None,
                 expect_bands: List[BandType] = None,
                 priority: Tuple[int, int] = None,
                 virtual: bool = False,
                 retryable: bool = True,
                 rerun_time: int = 0,
                 extra_config: dict = None):
        super().__init__(subtask_id=subtask_id,
                         subtask_name=subtask_name,
                         session_id=session_id,
                         task_id=task_id,
                         chunk_graph=chunk_graph,
                         expect_bands=expect_bands,
                         priority=priority,
                         virtual=virtual,
                         retryable=retryable,
                         rerun_time=rerun_time,
                         extra_config=extra_config)

    @property
    def expect_band(self):
        if self.expect_bands:
            return self.expect_bands[0]


class SubtaskResult(Serializable):
    subtask_id: str = StringField('subtask_id')
    session_id: str = StringField('session_id')
    task_id: str = StringField('task_id')
    status: SubtaskStatus = ReferenceField('status', SubtaskStatus)
    progress: float = Float64Field('progress', default=0.0)
    data_size: int = Int64Field('data_size', default=None)
    bands: List[BandType] = ListField('band', FieldTypes.tuple)
    error = AnyField('error', default=None)
    traceback = AnyField('traceback', default=None)

    def merge_bands(self, result: Optional['SubtaskResult']):
        if result and result.bands:
            bands = self.bands or []
            self.bands = sorted(set(bands + result.bands))
        return self


class SubtaskGraph(DAG, Iterable[Subtask]):
    """
    Subtask graph.
    """
    @classmethod
    def _extract_operands(cls, node: Subtask):
        from ...core.operand import Fetch, FetchShuffle
        for node in node.chunk_graph:
            if isinstance(node.op, (Fetch, FetchShuffle)):
                continue
            yield node.op



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures as futures
import itertools
import logging
from abc import ABC
from collections import namedtuple
from typing import Any, Callable, Coroutine, Dict, Type
from urllib.parse import urlparse

from ....serialization import serialize, deserialize
from ....utils import implements, classproperty
from ....utils import lazy_import
from ...debug import debug_async_timeout
from ...errors import ServerClosed
from ..communication.base import Channel, ChannelType, Server, Client
from ..communication.core import register_client, register_server
from ..communication.errors import ChannelClosed

ray = lazy_import("ray")
logger = logging.getLogger(__name__)

ChannelID = namedtuple("ChannelID", ["local_address", "client_id", "channel_index", "dest_address"])


class RayChannelException(Exception):

    def __init__(self, exc_type, exc_value: BaseException, exc_traceback):
        self.exc_type = exc_type
        self.exc_value = exc_value
        self.exc_traceback = exc_traceback


class RayChannelBase(Channel, ABC):
    """
    Channel for communications between ray processes.
    """
    __slots__ = '_channel_index', '_channel_id', '_in_queue', '_closed'

    name = 'ray'
    _channel_index_gen = itertools.count()

    def __init__(self,
                 local_address: str = None,
                 dest_address: str = None,
                 channel_index: int = None,
                 channel_id: ChannelID = None,
                 compression=None):
        super().__init__(local_address=local_address,
                         dest_address=dest_address,
                         compression=compression)
        self._channel_index = channel_index or next(self._channel_index_gen)
        self._channel_id = channel_id or ChannelID(
            local_address, _gen_client_id(), self._channel_index, dest_address)
        self._in_queue = asyncio.Queue()
        self._closed = asyncio.Event()

    @property
    def channel_id(self) -> ChannelID:
        return self._channel_id

    @property
    @implements(Channel.type)
    def type(self) -> ChannelType:
        return ChannelType.ray

    @implements(Channel.close)
    async def close(self):
        self._closed.set()

    @property
    @implements(Channel.closed)
    def closed(self) -> bool:
        return self._closed.is_set()


class RayClientChannel(RayChannelBase):
    """
    A channel from ray driver/actor to ray actor. Use ray call reply for client channel recv.
    """
    __slots__ = '_peer_actor',

    def __init__(self,
                 dest_address: str = None,
                 channel_index: int = None,
                 channel_id: ChannelID = None,
                 compression=None):
        super().__init__(None, dest_address, channel_index, channel_id, compression)
        # ray actor should be created with the address as the name.
        self._peer_actor: 'ray.actor.ActorHandle' = ray.get_actor(dest_address)

    @implements(Channel.send)
    async def send(self, message: Any):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed('Channel already closed, cannot send message')
        # Put ray object ref to queue
        await self._in_queue.put((message, self._peer_actor.__on_ray_recv__.remote(
            self.channel_id, serialize(message))))

    @implements(Channel.recv)
    async def recv(self):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed('Channel already closed, cannot recv message')
        try:
            # Wait on ray object ref
            message, object_ref = await self._in_queue.get()
            with debug_async_timeout('ray_object_retrieval_timeout', 'Client sent message is %s', message):
                result = await object_ref
            if isinstance(result, RayChannelException):
                raise result.exc_value.with_traceback(result.exc_traceback)
            return deserialize(*result)
        except ray.exceptions.RayActorError:
            if not self._closed.is_set():
                # raise a EOFError as the SocketChannel does
                raise EOFError('Server may be closed')
        except (RuntimeError, ServerClosed) as e:  # pragma: no cover
            if not self._closed.is_set():
                raise e


class RayServerChannel(RayChannelBase):
    """
    A channel from ray actor to ray driver/actor. Since ray actor can't call ray driver,
    we use ray call reply for server channel send. Note that there can't be multiple
    channel message sends for one received message, or else it will be taken as next
    message's reply.
    """
    __slots__ = '_out_queue', '_msg_recv_counter', '_msg_sent_counter'

    def __init__(self,
                 local_address: str = None,
                 channel_index: int = None,
                 channel_id: ChannelID = None,
                 compression=None):
        super().__init__(local_address, None, channel_index, channel_id, compression)
        self._out_queue = asyncio.Queue()
        self._msg_recv_counter = 0
        self._msg_sent_counter = 0

    @implements(Channel.send)
    async def send(self, message: Any):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed('Channel already closed, cannot send message')
        # Current process is ray actor, we use ray call reply to send message to ray driver/actor.
        # Not that we can only send once for every read message in channel, otherwise
        # it will be taken as other message's reply.
        await self._out_queue.put(serialize(message))
        self._msg_sent_counter += 1
        assert self._msg_sent_counter <= self._msg_recv_counter, \
            "RayServerChannel channel doesn't support send multiple replies for one message."

    @implements(Channel.recv)
    async def recv(self):
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed('Channel already closed, cannot write message')
        try:
            return deserialize(*(await self._in_queue.get()))
        except RuntimeError:  # pragma: no cover
            if not self._closed.is_set():
                raise

    async def __on_ray_recv__(self, message):
        """This method will be invoked when current process is a ray actor rather than a ray driver"""
        self._msg_recv_counter += 1
        await self._in_queue.put(message)
        # Avoid hang when channel is closed after `self._out_queue.get()` is awaited.
        done, _ = await asyncio.wait([self._out_queue.get(), self._closed.wait()],
                                     return_when=asyncio.FIRST_COMPLETED)
        if self._closed.is_set():  # pragma: no cover
            raise ChannelClosed('Channel already closed')
        if done:
            return await done.pop()


@register_server
class RayServer(Server):
    __slots__ = '_closed', '_channels', '_tasks'

    scheme = 'ray'
    _server_instance = None
    _ray_actor_started = False

    def __init__(self, address, channel_handler: Callable[[Channel], Coroutine] = None):
        super().__init__(address, channel_handler)
        self._closed = asyncio.Event()
        self._channels: Dict[ChannelID, RayServerChannel] = dict()
        self._tasks: Dict[ChannelID, asyncio.Task] = dict()

    @classproperty
    @implements(Server.client_type)
    def client_type(self) -> Type["Client"]:
        return RayClient

    @property
    @implements(Server.channel_type)
    def channel_type(self) -> ChannelType:
        return ChannelType.ray

    @classmethod
    def set_ray_actor_started(cls):
        cls._ray_actor_started = True

    @classmethod
    def is_ray_actor_started(cls):
        return cls._ray_actor_started

    @staticmethod
    @implements(Server.create)
    async def create(config: Dict) -> "RayServer":
        if not RayServer.is_ray_actor_started():
            logger.warning('Current process is not a ray actor, the ray server '
                           'will not receive messages from clients.')
        assert RayServer._server_instance is None
        config = config.copy()
        address = config.pop('address')
        handle_channel = config.pop('handle_channel')
        if urlparse(address).scheme != RayServer.scheme:  # pragma: no cover
            raise ValueError(f'Address for RayServer '
                             f'should be starts with "ray://", '
                             f'got {address}')
        if config:  # pragma: no cover
            raise TypeError(f'Creating RayServer got unexpected '
                            f'arguments: {",".join(config)}')
        server = RayServer(address, handle_channel)
        RayServer._server_instance = server
        return server

    @classmethod
    def get_instance(cls):
        return cls._server_instance

    @classmethod
    def clear(cls):
        cls._server_instance = None
        cls._ray_actor_started = False

    @implements(Server.start)
    async def start(self):
        # nothing needs to do for ray server
        pass

    @implements(Server.join)
    async def join(self, timeout=None):
        wait_coro = self._closed.wait()
        try:
            await asyncio.wait_for(wait_coro, timeout=timeout)
        except (futures.TimeoutError, asyncio.TimeoutError):  # pragma: no cover
            pass

    @implements(Server.on_connected)
    async def on_connected(self, *args, **kwargs):
        channel = args[0]
        assert isinstance(channel, RayServerChannel)
        if kwargs:  # pragma: no cover
            raise TypeError(f'{type(self).__name__} got unexpected '
                            f'arguments: {",".join(kwargs)}')
        await self.channel_handler(channel)

    @implements(Server.stop)
    async def stop(self):
        self._closed.set()
        for task in self._tasks.values():
            task.cancel()
        self._tasks = dict()
        for channel in self._channels.values():
            await channel.close()
        self._channels = dict()
        self.clear()

    @property
    @implements(Server.stopped)
    def stopped(self) -> bool:
        return self._closed.is_set()

    async def __on_ray_recv__(self, channel_id: ChannelID, message):
        if self.stopped:
            raise ServerClosed(f'Remote server {self.address} closed, but got message {deserialize(*message)} '
                               f'from channel {channel_id}')
        channel = self._channels.get(channel_id)
        if not channel:
            _, _, peer_channel_index, peer_dest_address = channel_id
            channel = RayServerChannel(peer_dest_address, peer_channel_index, channel_id)
            self._channels[channel_id] = channel
            self._tasks[channel_id] = asyncio.create_task(self.on_connected(channel))
        return await channel.__on_ray_recv__(message)


@register_client
class RayClient(Client):
    __slots__ = ()

    scheme = RayServer.scheme

    def __init__(self,
                 local_address: str,
                 dest_address: str,
                 channel: Channel):
        super().__init__(local_address,
                         dest_address,
                         channel)

    @staticmethod
    @implements(Client.connect)
    async def connect(dest_address: str,
                      local_address: str = None,
                      **kwargs) -> "Client":
        if urlparse(dest_address).scheme != RayServer.scheme:  # pragma: no cover
            raise ValueError(f'Destination address should start with "ray://" '
                             f'for RayClient, got {dest_address}')
        client_channel = RayClientChannel(dest_address)
        client = RayClient(local_address, dest_address, client_channel)
        return client

    @implements(Client.close)
    async def close(self):
        await super().close()


def _gen_client_id():
    import uuid
    return uuid.uuid4().hex



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from collections import defaultdict
from typing import Dict, Iterable, List, Tuple

from ....utils import ceildiv, lazy_import

ray = lazy_import('ray')
parallel_it = lazy_import('ray.util.iter')
ml_dataset = lazy_import('ray.util.data')


class ChunkRefBatch:
    def __init__(self,
                 shard_id: int,
                 obj_refs: 'ray.ObjectRef'):
        """Iterable batch holding a list of ray.ObjectRefs.

        Args:
            shard_id (int): id of the shard
            prefix (str): prefix name of the batch
            obj_refs (List[ray.ObjectRefs]): list of ray.ObjectRefs
        """
        self._shard_id = shard_id
        self._obj_refs = obj_refs

    @property
    def shard_id(self) -> int:
        return self._shard_id

    def __iter__(self) -> Iterable[pd.DataFrame]:
        """Returns the item_generator required from ParallelIteratorWorker."""
        for obj_ref in self._obj_refs:
            yield ray.get(obj_ref)


def _group_chunk_refs(chunk_addr_refs: List[Tuple[Tuple, 'ray.ObjectRef']],
                      num_shards: int):
    """Group fetched ray.ObjectRefs into a dict for later use.

    Args:
        chunk_addr_refs (List[Tuple[Tuple, ray.ObjectRef]]): a list of tuples of
            band & ray.ObjectRef of each chunk.
        num_shards (int): the number of shards that will be created for the MLDataset.

    Returns:
        Dict[str, List[ray.ObjectRef]]: a dict that defines which group of ray.ObjectRefs will
            be in an ChunkRefBatch.
    """
    group_to_obj_refs = defaultdict(list)
    if not num_shards:
        for addr, obj_ref in chunk_addr_refs:
            group_to_obj_refs[addr].append(obj_ref)
    else:
        splits = np.array_split([ref for _, ref in chunk_addr_refs],
                                num_shards)
        for idx, split in enumerate(splits):
            group_to_obj_refs['group-' + str(idx)] = list(split)
    return group_to_obj_refs


def _create_ml_dataset(name: str,
                       group_to_obj_refs: Dict[str, List['ray.ObjectRef']]):
    record_batches = []
    for rank, obj_refs in enumerate(group_to_obj_refs.values()):
        record_batches.append(ChunkRefBatch(shard_id=rank,
                                            obj_refs=obj_refs))
    worker_cls = ray.remote(num_cpus=0)(parallel_it.ParallelIteratorWorker)
    actors = [worker_cls.remote(g, False) for g in record_batches]
    it = parallel_it.from_actors(actors, name)
    ds = ml_dataset.from_parallel_iter(
        it, need_convert=False, batch_size=0, repeated=False)
    return ds


def _rechunk_if_needed(df, num_shards: int = None):
    chunk_size = df.extra_params.raw_chunk_size or max(df.shape)
    num_rows = df.shape[0]
    num_columns = df.shape[1]
    # if chunk size not set, num_chunks_in_row = 1
    # if chunk size is set more than max(df.shape), num_chunks_in_row = 1
    # otherwise, num_chunks_in_row depends on ceildiv(num_rows, chunk_size)
    num_chunks_in_row = ceildiv(num_rows, chunk_size)
    naive_num_partitions = ceildiv(num_rows, num_columns)

    need_re_execute = False
    # ensure each part holds all columns
    if chunk_size < num_columns:
        df = df.rebalance(axis=1, num_partitions=1)
        need_re_execute = True
    if num_shards and num_chunks_in_row < num_shards:
        df = df.rebalance(axis=0, num_partitions=num_shards)
        need_re_execute = True
    if not num_shards and num_chunks_in_row == 1:
        df = df.rebalance(axis=0, num_partitions=naive_num_partitions)
        need_re_execute = True
    if need_re_execute:
        df.execute()
    return df


def to_ray_mldataset(df,
                     num_shards: int = None):
    """Create a MLDataset from Mars DataFrame

    Args:
        df (mars.dataframe.Dataframe): the Mars DataFrame
        num_shards (int, optional): the number of shards that will be created
            for the MLDataset. Defaults to None.
            If num_shards is None, chunks will be grouped by nodes where they lie.
            Otherwise, chunks will be grouped by their order in DataFrame.

    Returns:
        a MLDataset
    """
    df = _rechunk_if_needed(df, num_shards)
    # chunk_addr_refs is fetched directly rather than in batches
    # during `fetch` procedure, it'll be checked that df has been executed
    # items in chunk_addr_refs are ordered by positions in df
    # while adjacent chunks may belong to different addrs, i.e.
    #       chunk1 for addr1,
    #       chunk2 & chunk3 for addr2,
    #       chunk4 for addr1
    fetched_infos: Dict[str, List] = df.fetch_infos(fields=['band', 'object_id'])
    chunk_addr_refs: List[Tuple[Tuple, 'ray.ObjectRef']] = [(band, object_id) for band, object_id in
                                                            zip(fetched_infos['band'],
                                                                fetched_infos['object_id'])]
    group_to_obj_refs: Dict[str, List[ray.ObjectRef]] = _group_chunk_refs(chunk_addr_refs, num_shards)
    return _create_ml_dataset("from_mars", group_to_obj_refs)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

import numpy as np
try:
    import scipy.sparse as sps
except ImportError:  # pragma: no cover
    sps = None

from .core import Serializer, buffered, serialize, deserialize


class CsrMatrixSerializer(Serializer):
    serializer_name = 'sps.csr_matrix'

    @buffered
    def serialize(self, obj: Any, context: Dict):
        data_header, data_buffers = serialize(obj.data)
        idx_header, idx_buffers = serialize(obj.indices)
        indptr_header, indptr_buffers = serialize(obj.indptr)
        header = {
            'data_header': data_header, 'data_buf_num': len(data_buffers),
            'idx_header': idx_header, 'idx_buf_num': len(idx_buffers),
            'indptr_header': indptr_header,
            'shape': list(obj.shape),
        }
        return header, data_buffers + idx_buffers + indptr_buffers

    def deserialize(self, header: Dict, buffers: List, context: Dict):
        data_buf_num = header['data_buf_num']
        idx_buf_num = header['idx_buf_num']
        data_buffers = buffers[:data_buf_num]
        idx_buffers = buffers[data_buf_num:data_buf_num + idx_buf_num]
        indptr_buffers = buffers[data_buf_num + idx_buf_num:]

        data = deserialize(header['data_header'], data_buffers)
        indices = deserialize(header['idx_header'], idx_buffers)
        indptr = deserialize(header['indptr_header'], indptr_buffers)
        shape = tuple(header['shape'])

        empty_arr = np.zeros(0, dtype=data.dtype)

        target_csr = sps.coo_matrix((empty_arr, (empty_arr,) * 2), dtype=data.dtype,
                                    shape=shape).tocsr()
        target_csr.data, target_csr.indices, target_csr.indptr = data, indices, indptr
        return target_csr


if sps:  # pragma: no branch
    CsrMatrixSerializer.register(sps.csr_matrix)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from enum import Enum
from string import ascii_letters, digits
from typing import Any, Optional

from ...core import TileableGraph
from ...serialization.serializables import Serializable, StringField, \
    ReferenceField, Int32Field, BoolField, AnyField, DictField, \
    Float64Field


class TaskStatus(Enum):
    pending = 0
    running = 1
    terminated = 2


class Task(Serializable):
    task_id: str = StringField('task_id')
    task_name: str = StringField('task_name')
    session_id: str = StringField('session_id')
    parent_task_id: str = StringField('parent_task_id')
    tileable_graph: TileableGraph = ReferenceField(
        'tileable_graph', TileableGraph)
    fuse_enabled: bool = BoolField('fuse_enabled')
    rerun_time: int = Int32Field('rerun_time')
    extra_config: dict = DictField('extra_config')

    def __init__(self,
                 task_id: str = None,
                 session_id: str = None,
                 tileable_graph: TileableGraph = None,
                 task_name: str = None,
                 parent_task_id: str = None,
                 fuse_enabled: bool = True,
                 rerun_time: int = 0,
                 extra_config: dict = None):
        super().__init__(task_id=task_id, task_name=task_name,
                         session_id=session_id,
                         parent_task_id=parent_task_id,
                         tileable_graph=tileable_graph,
                         fuse_enabled=fuse_enabled,
                         rerun_time=rerun_time,
                         extra_config=extra_config)


class TaskResult(Serializable):
    task_id: str = StringField('task_id')
    session_id: str = StringField('session_id')
    stage_id: str = StringField('stage_id')
    start_time: Optional[float] = Float64Field('start_time')
    end_time: Optional[float] = Float64Field('end_time')
    progress: Optional[float] = Float64Field('progress')
    status: TaskStatus = ReferenceField('status', TaskStatus)
    error = AnyField('error')
    traceback = AnyField('traceback')

    def __init__(self,
                 task_id: str = None,
                 session_id: str = None,
                 stage_id: str = None,
                 start_time: Optional[float] = None,
                 end_time: Optional[float] = None,
                 progress: Optional[float] = None,
                 status: TaskStatus = None,
                 error: Any = None,
                 traceback: Any = None):
        super().__init__(task_id=task_id,
                         session_id=session_id,
                         stage_id=stage_id,
                         start_time=start_time,
                         end_time=end_time,
                         progress=progress,
                         status=status,
                         error=error,
                         traceback=traceback)


def new_task_id():
    return ''.join(random.choice(ascii_letters + digits) for _ in range(24))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import List, Optional, Tuple, Type, TypeVar, Union

from ... import oscar as mo
from ...lib.aio import alru_cache
from ..subtask import Subtask

APIType = TypeVar('APIType', bound='SchedulingAPI')


class SchedulingAPI(ABC):
    def __init__(self, session_id: str, address: str,
                 manager_ref=None, queueing_ref=None):
        self._session_id = session_id
        self._address = address

        self._manager_ref = manager_ref
        self._queueing_ref = queueing_ref

    @classmethod
    @alru_cache
    async def create(cls: Type[APIType],
                     session_id: str,
                     address: str) -> APIType:
        from .supervisor.manager import SubtaskManagerActor
        manager_ref = await mo.actor_ref(
            SubtaskManagerActor.gen_uid(session_id), address=address
        )
        from .supervisor.queueing import SubtaskQueueingActor
        queueing_ref = await mo.actor_ref(
            SubtaskQueueingActor.gen_uid(session_id), address=address
        )

        scheduling_api = SchedulingAPI(
            session_id, address, manager_ref, queueing_ref)
        return scheduling_api

    async def add_subtasks(self,
                           subtasks: List[Subtask],
                           priorities: Optional[List[Tuple]] = None):
        """
        Submit subtasks into scheduling service

        Parameters
        ----------
        subtasks
            list of subtasks to be submitted to service
        priorities
            list of priorities of subtasks
        """
        if priorities is None:
            priorities = [subtask.priority or tuple() for subtask in subtasks]
        await self._manager_ref.add_subtasks(subtasks, priorities)

    @mo.extensible
    async def update_subtask_priority(self,
                                      subtask_id: str,
                                      priority: Tuple):
        """
        Update priorities of subtasks

        Parameters
        ----------
        subtask_id
            id of subtask to update priority
        priority
            list of priority of subtasks
        """
        raise NotImplementedError

    @update_subtask_priority.batch
    async def update_subtask_priority(self, args_list, kwargs_list):
        await self._queueing_ref.update_subtask_priority.batch(
            *(self._queueing_ref.update_subtask_priority.delay(*args, **kwargs)
              for args, kwargs in zip(args_list, kwargs_list)))

    async def cancel_subtasks(self,
                              subtask_ids: List[str],
                              kill_timeout: Union[float, int] = 5):
        """
        Cancel pending and running subtasks.

        Parameters
        ----------
        subtask_ids
            ids of subtasks to cancel
        kill_timeout
            timeout seconds to kill actor process forcibly
        """
        await self._manager_ref.cancel_subtasks(
            subtask_ids, kill_timeout=kill_timeout)

    async def finish_subtasks(self,
                              subtask_ids: List[str],
                              schedule_next: bool = True):
        """
        Mark subtasks as finished, letting scheduling service to schedule
        next tasks in the ready queue

        Parameters
        ----------
        subtask_ids
            ids of subtasks to mark as finished
        schedule_next
            whether to schedule succeeding subtasks
        """
        await self._manager_ref.finish_subtasks(subtask_ids, schedule_next)


class MockSchedulingAPI(SchedulingAPI):
    @classmethod
    async def create(cls: Type[APIType],
                     session_id: str,
                     address: str) -> APIType:
        from .supervisor import GlobalSlotManagerActor, AutoscalerActor
        await mo.create_actor(GlobalSlotManagerActor,
                              uid=GlobalSlotManagerActor.default_uid(),
                              address=address)
        await mo.create_actor(AutoscalerActor, {},
                              uid=AutoscalerActor.default_uid(),
                              address=address)

        from ... import resource as mars_resource
        from .worker import SubtaskExecutionActor, \
            WorkerSlotManagerActor, WorkerQuotaManagerActor
        await mo.create_actor(SubtaskExecutionActor,
                              subtask_max_retries=0,
                              uid=SubtaskExecutionActor.default_uid(),
                              address=address)
        await mo.create_actor(WorkerSlotManagerActor,
                              uid=WorkerSlotManagerActor.default_uid(),
                              address=address)
        await mo.create_actor(WorkerQuotaManagerActor,
                              {'quota_size': mars_resource.virtual_memory().total},
                              uid=WorkerQuotaManagerActor.default_uid(),
                              address=address)

        from .supervisor import SchedulingSupervisorService
        service = SchedulingSupervisorService({}, address)
        await service.create_session(session_id)
        return await super().create(session_id, address)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ...config import options
from ...core import OutputType
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameSum(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = opcodes.SUM
    _func_name = 'sum'

    @property
    def is_atomic(self):
        return self.min_count == 0

    @classmethod
    def get_reduction_callable(cls, op):
        from .aggregation import where_function
        skipna, min_count = op.skipna, op.min_count

        def sum_(value):
            if min_count == 0:
                return value.sum(skipna=skipna)
            else:
                return where_function(value.count() >= min_count, value.sum(skipna=skipna), np.nan)

        return sum_


def sum_series(df, axis=None, skipna=None, level=None, min_count=0,
               combine_size=None, method=None):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameSum(axis=axis, skipna=skipna, level=level, min_count=min_count,
                      combine_size=combine_size, output_types=[OutputType.scalar],
                      use_inf_as_na=use_inf_as_na, method=method)
    return op(df)


def sum_dataframe(df, axis=None, skipna=None, level=None, min_count=0, numeric_only=None,
                  combine_size=None, method=None):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameSum(axis=axis, skipna=skipna, level=level, min_count=min_count,
                      numeric_only=numeric_only, combine_size=combine_size,
                      output_types=[OutputType.series], use_inf_as_na=use_inf_as_na,
                      method=method)
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures as futures
import os
import socket
import sys
import tempfile
from abc import ABCMeta
from asyncio import StreamReader, StreamWriter, AbstractServer
from functools import lru_cache
from hashlib import md5
from typing import Any, Dict, Callable, Coroutine, Type
from urllib.parse import urlparse

from ....serialization import AioSerializer, AioDeserializer, deserialize
from ....utils import implements, to_binary, classproperty
from .base import Channel, ChannelType, Server, Client
from .core import register_client, register_server
from .utils import read_buffers, write_buffers

_is_windows: bool = sys.platform.startswith('win')


class SocketChannel(Channel):
    __slots__ = 'reader', 'writer', '_channel_type', '_send_lock', '_recv_lock'

    name = 'socket'

    def __init__(self,
                 reader: StreamReader,
                 writer: StreamWriter,
                 local_address: str = None,
                 dest_address: str = None,
                 compression: int = None,
                 channel_type: ChannelType = None):
        super().__init__(local_address=local_address,
                         dest_address=dest_address,
                         compression=compression)
        self.reader = reader
        self.writer = writer
        self._channel_type = channel_type

        self._send_lock = asyncio.Lock()
        self._recv_lock = asyncio.Lock()

    @property
    @implements(Channel.type)
    def type(self) -> ChannelType:
        return self._channel_type

    @implements(Channel.send)
    async def send(self, message: Any):
        # get buffers
        compress = self.compression or 0
        serializer = AioSerializer(message, compress=compress)
        buffers = await serializer.run()

        # write buffers
        write_buffers(self.writer, buffers)
        async with self._send_lock:
            # add lock, or when parallel send,
            # assertion error may be raised
            await self.writer.drain()

    @implements(Channel.recv)
    async def recv(self):
        deserializer = AioDeserializer(self.reader)
        async with self._recv_lock:
            header = await deserializer.get_header()
            buffers = await read_buffers(header, self.reader)
        return deserialize(header, buffers)

    @implements(Channel.close)
    async def close(self):
        self.writer.close()
        try:
            await self.writer.wait_closed()
        except ConnectionResetError:  # pragma: no cover
            pass

    @property
    @implements(Channel.closed)
    def closed(self):
        return self.writer.is_closing()


class _BaseSocketServer(Server, metaclass=ABCMeta):
    __slots__ = '_aio_server', '_channels'

    def __init__(self,
                 address: str,
                 aio_server: AbstractServer,
                 channel_handler: Callable[[Channel], Coroutine] = None):
        super().__init__(address, channel_handler)
        # asyncio.Server
        self._aio_server = aio_server
        self._channels = []

    @implements(Server.start)
    async def start(self):
        await self._aio_server.start_serving()

    @implements(Server.join)
    async def join(self, timeout=None):
        if timeout is None:
            await self._aio_server.serve_forever()
        else:
            future = asyncio.create_task(self._aio_server.serve_forever())
            try:
                await asyncio.wait_for(future, timeout=timeout)
            except (futures.TimeoutError, asyncio.TimeoutError):
                future.cancel()

    @implements(Server.on_connected)
    async def on_connected(self, *args, **kwargs):
        reader, writer = args
        local_address = kwargs.pop('local_address', None)
        dest_address = kwargs.pop('dest_address', None)
        if kwargs:  # pragma: no cover
            raise TypeError(f'{type(self).__name__} got unexpected '
                            f'arguments: {",".join(kwargs)}')
        channel = SocketChannel(reader, writer,
                                local_address=local_address,
                                dest_address=dest_address,
                                channel_type=self.channel_type)
        self._channels.append(channel)
        # handle over channel to some handlers
        await self.channel_handler(channel)

    @implements(Server.stop)
    async def stop(self):
        self._aio_server.close()
        await self._aio_server.wait_closed()
        # close all channels
        await asyncio.gather(
            *(channel.close() for channel in self._channels
              if not channel.closed))

    @property
    @implements(Server.stopped)
    def stopped(self) -> bool:
        return not self._aio_server.is_serving()


@register_server
class SocketServer(_BaseSocketServer):
    __slots__ = 'host', 'port'

    scheme = None

    def __init__(self,
                 host: str,
                 port: int,
                 aio_server: AbstractServer,
                 channel_handler: Callable[[Channel], Coroutine] = None):
        address = f'{host}:{port}'
        super().__init__(address, aio_server,
                         channel_handler=channel_handler)
        self.host = host
        self.port = port

    @classproperty
    @implements(Server.client_type)
    def client_type(self) -> Type["Client"]:
        return SocketClient

    @property
    @implements(Server.channel_type)
    def channel_type(self) -> ChannelType:
        return ChannelType.remote

    @staticmethod
    @implements(Server.create)
    async def create(config: Dict) -> "Server":
        config = config.copy()
        if 'address' in config:
            address = config.pop('address')
            host, port = address.split(':', 1)
            port = int(port)
        else:
            host = config.pop('host')
            port = int(config.pop('port'))
        handle_channel = config.pop('handle_channel')
        if 'start_serving' not in config:
            config['start_serving'] = False

        async def handle_connection(reader, writer):
            # create a channel when client connected
            return await server.on_connected(reader, writer,
                                             local_address=server.address)

        aio_server = await asyncio.start_server(
            handle_connection, host=host, port=port, **config)

        if _is_windows:
            for sock in aio_server.sockets:
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)

        server = SocketServer(host, port, aio_server,
                              channel_handler=handle_channel)
        return server


@register_client
class SocketClient(Client):
    __slots__ = ()

    scheme = SocketServer.scheme

    @staticmethod
    @implements(Client.connect)
    async def connect(dest_address: str,
                      local_address: str = None,
                      **kwargs) -> "Client":
        host, port = dest_address.split(':', 1)
        port = int(port)
        (reader, writer) = await asyncio.open_connection(
            host=host, port=port, **kwargs)
        channel = SocketChannel(reader, writer,
                                local_address=local_address,
                                dest_address=dest_address)
        return SocketClient(local_address, dest_address, channel)


TEMPDIR = tempfile.gettempdir()


@lru_cache(100)
def _gen_unix_socket_default_path(process_index):
    return f'{TEMPDIR}/mars/' \
           f'{md5(to_binary(str(process_index))).hexdigest()}'  # nosec


@register_server
class UnixSocketServer(_BaseSocketServer):
    __slots__ = 'process_index', 'path'

    scheme = 'unixsocket'

    def __init__(self,
                 process_index: int,
                 aio_server: AbstractServer,
                 path: str,
                 channel_handler: Callable[[Channel], Coroutine] = None):
        address = f'{self.scheme}:///{process_index}'
        super().__init__(address, aio_server,
                         channel_handler=channel_handler)
        self.process_index = process_index
        self.path = path

    @classproperty
    @implements(Server.client_type)
    def client_type(self) -> Type["Client"]:
        return UnixSocketClient

    @property
    @implements(Server.channel_type)
    def channel_type(self) -> ChannelType:
        return ChannelType.ipc

    @staticmethod
    @implements(Server.create)
    async def create(config: Dict) -> "Server":
        config = config.copy()
        if 'address' in config:
            process_index = int(urlparse(config.pop('address')).path.lstrip('/'))
        else:
            process_index = config.pop('process_index')
        handle_channel = config.pop('handle_channel')
        path = config.pop('path', _gen_unix_socket_default_path(process_index))

        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname, exist_ok=True)

        if 'start_serving' not in config:
            config['start_serving'] = False

        async def handle_connection(reader, writer):
            # create a channel when client connected
            return await server.on_connected(reader, writer,
                                             local_address=server.address)

        aio_server = await asyncio.start_unix_server(
            handle_connection, path=path, **config)

        for sock in aio_server.sockets:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)

        server = UnixSocketServer(process_index, aio_server, path,
                                  channel_handler=handle_channel)
        return server

    @implements(Server.stop)
    async def stop(self):
        await super().stop()
        try:
            os.remove(self.path)
        except OSError:  # pragma: no cover
            pass


@register_client
class UnixSocketClient(Client):
    __slots__ = ()

    scheme = UnixSocketServer.scheme

    @staticmethod
    @lru_cache(100)
    def _get_process_index(addr):
        return int(urlparse(addr).path.lstrip('/'))

    @staticmethod
    @implements(Client.connect)
    async def connect(dest_address: str,
                      local_address: str = None,
                      **kwargs) -> "Client":
        process_index = UnixSocketClient._get_process_index(dest_address)
        path = kwargs.pop('path',
                          _gen_unix_socket_default_path(process_index))
        try:
            (reader, writer) = await asyncio.open_unix_connection(path, **kwargs)
        except FileNotFoundError:
            raise ConnectionRefusedError('Cannot connect unix socket '
                                         'due to file not exists')
        channel = SocketChannel(reader, writer,
                                local_address=local_address,
                                dest_address=dest_address)
        return UnixSocketClient(local_address, dest_address, channel)
